//
//  BannerView.swift
//  MiniScanner
//
//  Created by George Makhoul on 11/06/2024.
//  Copyright © 2024 AppsNectar. All rights reserved.
//

import UIKit

protocol BannerViewDelegate {
    func updateSelectedIndex(index: Int)
    func delete(index: Int)
    func preview(index: Int)
    func update(session: MultiPageScanSession?)
}

class BannerView: UIView {
    // MARK: - Outlets
    @IBOutlet var view: UIView!
    @IBOutlet weak var pagerView: FSPagerView!{
        didSet {
            self.pagerView.register(FSPagerViewCell.self, forCellWithReuseIdentifier: "cell")
            self.typeIndex = 3
            pagerView.dataSource = self
            pagerView.delegate = self
            
        }
    }
    
    var signatureImage: UIImage!
    var signAllDoc: Bool = false
    var donePressed: Bool = false
    var isEditing: Bool = false
    fileprivate let transformerTypes: [FSPagerViewTransformerType] = [.crossFading,
                                                                      .zoomOut,
                                                                      .depth,
                                                                      .linear,
                                                                      .overlap,
                                                                      .ferrisWheel,
                                                                      .invertedFerrisWheel,
                                                                      .coverFlow,
                                                                      .cubic]
    var options:ImageScannerOptions!
    
    fileprivate var typeIndex = 0 {
        didSet {
            let type = self.transformerTypes[typeIndex]
            self.pagerView.transformer = FSPagerViewTransformer(type:type)
            switch type {
            case .crossFading, .zoomOut, .depth:
                self.pagerView.itemSize = FSPagerView.automaticSize
                self.pagerView.decelerationDistance = 1
            case .linear, .overlap:
                let transform = CGAffineTransform(scaleX: 0.8, y: 0.75)
                self.pagerView.itemSize = self.pagerView.frame.size.applying(transform)
                self.pagerView.decelerationDistance = FSPagerView.automaticDistance
            case .ferrisWheel, .invertedFerrisWheel:
                self.pagerView.itemSize = CGSize(width: 180, height: 140)
                self.pagerView.decelerationDistance = FSPagerView.automaticDistance
            case .coverFlow:
                self.pagerView.itemSize = CGSize(width: 220, height: 170)
                self.pagerView.decelerationDistance = FSPagerView.automaticDistance
            case .cubic:
                let transform = CGAffineTransform(scaleX: 0.9, y: 0.9)
                self.pagerView.itemSize = self.pagerView.frame.size.applying(transform)
                self.pagerView.decelerationDistance = 1
            }
        }
    }
    var delegate: BannerViewDelegate?
    
    // MARK: - Properties
    
    var scanSession: MultiPageScanSession?
    
    
    // MARK: - Init
    override init(frame: CGRect) {
        super.init(frame: frame)
        setupNib()
    }
    
    required init?(coder aDecoder: NSCoder) {
        super.init(coder: aDecoder)
        
        setupNib()
    }
    
    private func setupNib() {
        Bundle.main.loadNibNamed("BannerView", owner: self, options: nil)
        
        view.frame = bounds
        view.autoresizingMask = [.flexibleHeight, .flexibleWidth]
        
        addSubview(view)
        clearUI()
    }
    
    override func layoutSubviews() {
        super.layoutSubviews()
        
        clearUI()
    }
    
    // MARK: - Methods
    private func clearUI() {
        view.backgroundColor = .clear
        self.options = ImageScannerOptions()
    }
    
    
    
    func set(session: MultiPageScanSession) {
        self.scanSession = session
        clearUI()
    }
    func reloadSignature(with : UIImage, isAll: Bool = false) {
        self.signatureImage = with
        self.signAllDoc = isAll
        pagerView.reloadData()
    }
    
    func saveSignarure() {
        donePressed = true
        let cell = pagerView.cellForItem(at: pagerView.currentIndex)
        cell?.selectedStickerView1?.showEditing = false
        let image =  cell?.contentView.TakeScreenshot()
        cell?.selectedStickerView1?.removeFromSuperview()
        self.scanSession?.remove(index: pagerView.currentIndex)
        
        let scannedItem = ScannedItem(originalImage: image ?? UIImage(), quad: nil)
        scannedItem.colorOption = self.options.defaultColorRenderOption
        scannedItem.rotation = 180
        scannedItem.render { (_) in }
        self.scanSession?.add(item: scannedItem, at: pagerView.currentIndex)
        cell?.imageView?.image = image
        delegate?.update(session: self.scanSession)
    }
    
}


extension BannerView: FSPagerViewDataSource,FSPagerViewDelegate {
    func numberOfItems(in pagerView: FSPagerView) -> Int {
        scanSession?.scannedItems.count ?? 0
    }
    func pagerView(_ pagerView: FSPagerView, cellForItemAt index: Int) -> FSPagerViewCell {
        let cell = pagerView.dequeueReusableCell(withReuseIdentifier: "cell", at: index)
        cell.index = index
        cell.delegate = self
        
        if let item = scanSession?.scannedItems[index] {
            cell.imageView?.image = item.renderedImage?.retrieveImage()
        }
        if isEditing {
            if signAllDoc {
                if cell.sign == nil {
                    cell.sign = signatureImage
                }
            } else {
                if index == pagerView.currentIndex {
                    if cell.sign == nil {
                        cell.sign = signatureImage
                    }
                }
            }
        }
        
        return cell
    }
    
    func pagerViewWillEndDragging(_ pagerView: FSPagerView, targetIndex: Int) {
        delegate?.updateSelectedIndex(index: targetIndex)
    }
}

extension BannerView: FSPagerViewCellDelegate {
    func preview(at: Int) {
        delegate?.preview(index: at)
    }
    
    func delete(at: Int) {
        delegate?.delete(index: at)
    }
}
